from copy import deepcopy

from apronpy.coeff import PyMPQScalarCoeff
from apronpy.environment import PyEnvironment
from apronpy.lincons0 import ConsTyp
from apronpy.polka import PyPolkaMPQstrict
from apronpy.t1p import PyT1pMPQ
from apronpy.tcons1 import PyTcons1, PyTcons1Array
from apronpy.texpr0 import TexprOp, TexprRtype, TexprRdir
from apronpy.texpr1 import PyTexpr1
from apronpy.var import PyVar


def do():
    size = 1#000000

    for i in range(size):

        x = PyVar('x')
        y = PyVar('y')
        r_vars = [x, y]
        for i in range(10, 52):
            name = 'x{}'.format(i)
            r_vars.append(PyVar(name))
        e = PyEnvironment([], r_vars)
        p0 = [PyT1pMPQ(e)]

        # o1 = PyVar('o1')
        # o2 = PyVar('o2')
        # h1 = PyVar('h1')
        # h2 = PyVar('h2')
        # h3 = PyVar('h3')
        # h4 = PyVar('h4')
        # h5 = PyVar('h5')
        # h6 = PyVar('h6')
        # h7 = PyVar('h7')
        # h8 = PyVar('h8')
        # h9 = PyVar('h9')
        # e = PyEnvironment([], [o1, o2, h1, h2, h3, h4, h5, h6, h7, h8, h9])
        #
        #
        o01 = PyTexpr1.var(e, PyVar('x'))
        o02 = PyTexpr1.cst(e, PyMPQScalarCoeff(0))
        x0 = PyTexpr1.binop(TexprOp.AP_TEXPR_SUB, o02, o01, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        c0 = PyTcons1.make(x0, ConsTyp.AP_CONS_SUP)     # o2 - o1 > 0
        a0 = PyTcons1Array([c0])
        p1 = [deepcopy(p).meet(a0) for p in p0]
        print(p1)

        # x1a = PyTexpr1.var(e, PyVar('o1'))
        # p2a = [deepcopy(p).substitute(PyVar('o1'), x1a) for p in p1]     # o1 = o1
        # o1a = PyTexpr1.var(e, PyVar('o1'))
        # c1a = PyTcons1.make(o1a, ConsTyp.AP_CONS_SUPEQ)     # o1 >= 0
        # a1a = PyTcons1Array([c1a])
        # p3a = [deepcopy(p).meet(a1a) for p in p2a]
        # #
        # x1b = PyTexpr1.cst(e, PyMPQScalarCoeff(0.0))
        # p2b = [deepcopy(p).substitute(PyVar('o1'), x1b) for p in p1]     # o1 = 0
        # o2b = PyTexpr1.var(e, PyVar('o2'))
        # k2b = PyTexpr1.cst(e, PyMPQScalarCoeff(-1.0))
        # t2b = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, k2b, o2b, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # c1b = PyTcons1.make(t2b, ConsTyp.AP_CONS_SUP)
        # a1b = PyTcons1Array([c1b])
        # p3b = [deepcopy(p).meet(a1b) for p in p2b]
        # #
        # p3 = p3a + p3b
        #
        # h21 = PyTexpr1.var(e, PyVar('h1'))
        # h22 = PyTexpr1.var(e, PyVar('h2'))
        # h23 = PyTexpr1.var(e, PyVar('h3'))
        # h24 = PyTexpr1.var(e, PyVar('h4'))
        # h25 = PyTexpr1.var(e, PyVar('h5'))
        # h26 = PyTexpr1.var(e, PyVar('h6'))
        # h27 = PyTexpr1.var(e, PyVar('h7'))
        # h28 = PyTexpr1.var(e, PyVar('h8'))
        # h29 = PyTexpr1.var(e, PyVar('h9'))
        # k20 = PyTexpr1.cst(e, PyMPQScalarCoeff(-0.2))
        # t20 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, k20, h21, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # k21 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.3))
        # t21 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, k21, h22, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # k22 = PyTexpr1.cst(e, PyMPQScalarCoeff(-0.1))
        # t22 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, k22, h23, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # k23 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.2))
        # t23 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, k23, h24, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # k24 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.1))
        # t24 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, k24, h25, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # k25 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.3))
        # t25 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, k25, h26, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # k26 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.4))
        # t26 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, k26, h27, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # k27 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.2))
        # t27 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, k27, h28, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # k28 = PyTexpr1.cst(e, PyMPQScalarCoeff(-0.1))
        # t28 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, k28, h29, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # k29 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.4))
        # x20 = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, t20, t21, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # x21 = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, x20, t22, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # x22 = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, x21, t23, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # x23 = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, x22, t24, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # x24 = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, x23, t25, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # x25 = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, x24, t26, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # x26 = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, x25, t27, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # x27 = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, x26, t28, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # x28 = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, x27, k29, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # # r = repr(x28)
        # p4 = [deepcopy(p).substitute(PyVar('o2'), x28) for p in p3]
        #
        # h31 = PyTexpr1.var(e, PyVar('h1'))
        # h32 = PyTexpr1.var(e, PyVar('h3'))
        # h33 = PyTexpr1.var(e, PyVar('h3'))
        # h34 = PyTexpr1.var(e, PyVar('h4'))
        # h35 = PyTexpr1.var(e, PyVar('h5'))
        # h36 = PyTexpr1.var(e, PyVar('h6'))
        # h37 = PyTexpr1.var(e, PyVar('h7'))
        # h38 = PyTexpr1.var(e, PyVar('h8'))
        # h39 = PyTexpr1.var(e, PyVar('h9'))
        # k30 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.4))
        # t30 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, k30, h31, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # k31 = PyTexpr1.cst(e, PyMPQScalarCoeff(-0.1))
        # t31 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, k31, h32, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # k32 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.2))
        # t32 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, k32, h33, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # k33 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.3))
        # t33 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, k33, h34, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # k34 = PyTexpr1.cst(e, PyMPQScalarCoeff(-0.1))
        # t34 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, k34, h35, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # k35 = PyTexpr1.cst(e, PyMPQScalarCoeff(-0.2))
        # t35 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, k35, h36, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # k36 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.4))
        # t36 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, k36, h37, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # k37 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.3))
        # t37 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, k37, h38, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # k38 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.1))
        # t38 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, k38, h39, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # k39 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.4))
        # x30 = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, t30, t31, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # x31 = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, x30, t32, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # x32 = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, x31, t33, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # x33 = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, x32, t34, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # x34 = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, x33, t35, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # x35 = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, x34, t36, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # x36 = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, x35, t37, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # x37 = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, x36, t38, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # x38 = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, x37, k39, TexprRtype.AP_RTYPE_DOUBLE, TexprRdir.AP_RDIR_RND)
        # p5 = [deepcopy(p).substitute(PyVar('o1'), x38) for p in p4]
        # print(p5)


do()



"""
APRON Coefficients
==================

:Author: Caterina Urban
"""
from _ctypes import Union, Structure, POINTER, byref
from abc import ABCMeta
from copy import deepcopy
from ctypes import c_uint
from enum import IntEnum

from apronpy.cdll import libapron
from apronpy.interval import Interval, PyInterval, PyDoubleInterval, PyMPQInterval, PyMPFRInterval
from apronpy.mpfr import Rnd
from apronpy.scalar import Scalar, PyScalar, PyDoubleScalar, PyMPQScalar, PyMPFRScalar


class CoeffDiscr(IntEnum):
    """
    typedef enum ap_coeff_discr_t {
      AP_COEFF_SCALAR,
      AP_COEFF_INTERVAL
    } ap_coeff_discr_t;
    """
    AP_COEFF_SCALAR = 0
    AP_COEFF_INTERVAL = 1


class Coeff(Structure):
    """
    typedef struct ap_coeff_t {
      ap_coeff_discr_t discr; /* discriminant for coefficient */
      union {
        ap_scalar_t* scalar;       /* cst (normal linear expression) */
        ap_interval_t* interval;   /* interval (quasi-linear expression) */
      } val;
    } ap_coeff_t;
    """
    class Val(Union):
        """
        union {
          ap_scalar_t* scalar;       /* cst (normal linear expression) */
          ap_interval_t* interval;   /* interval (quasi-linear expression) */
        } val;
        """

        _fields_ = [
            ('scalar', POINTER(Scalar)),
            ('interval', POINTER(Interval))
        ]

    _fields_ = [
        ('discr', c_uint),
        ('val', Val)
    ]

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        result = libapron.ap_coeff_alloc(self.discr)
        libapron.ap_coeff_set(result, byref(self))
        memodict[id(self)] = result.contents
        return result.contents

    def __repr__(self):
        if self.discr == CoeffDiscr.AP_COEFF_INTERVAL:
            return '{}'.format(self.val.interval.contents)
        else:
            assert self.discr == CoeffDiscr.AP_COEFF_SCALAR
            return '{}'.format(self.val.scalar.contents)


class PyCoeff(metaclass=ABCMeta):

    def __init__(self, value, discr: CoeffDiscr = CoeffDiscr.AP_COEFF_SCALAR):
        if isinstance(value, Coeff):
            self.coeff = byref(value)
        else:
            self.coeff = libapron.ap_coeff_alloc(discr)
            if discr == CoeffDiscr.AP_COEFF_INTERVAL:
                libapron.ap_coeff_set_interval(self, value)
            else:
                assert discr == CoeffDiscr.AP_COEFF_SCALAR
                libapron.ap_coeff_set_scalar(self, value)

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        result = type(self)(deepcopy(self.coeff.contents))
        memodict[id(self)] = result
        return result

    def __del__(self):
        libapron.ap_coeff_free(self)
        del self.coeff

    @property
    def _as_parameter_(self):
        return self.coeff

    @staticmethod
    def from_param(argument):
        assert isinstance(argument, PyCoeff)
        return argument

    def __repr__(self):
        return '{}'.format(self.coeff.contents)

    def __lt__(self, other: 'PyCoeff'):
        assert isinstance(other, PyCoeff)
        return libapron.ap_coeff_cmp(self, other) == -1

    def __le__(self, other: 'PyCoeff'):
        assert isinstance(other, PyCoeff)
        return self.__lt__(other) or self.__eq__(other)

    def __eq__(self, other: 'PyCoeff'):
        assert isinstance(other, PyCoeff)
        return libapron.ap_coeff_cmp(self, other) == 0

    def __ne__(self, other: 'PyCoeff'):
        assert isinstance(other, PyCoeff)
        return not self.__eq__(other)

    def __ge__(self, other: 'PyCoeff'):
        assert isinstance(other, PyCoeff)
        return self.__gt__(other) or self.__eq__(other)

    def __gt__(self, other: 'PyCoeff'):
        assert isinstance(other, PyCoeff)
        return libapron.ap_coeff_cmp(self, other) == 1

    def __neg__(self) -> 'PyCoeff':
        if self.coeff.contents.discr == CoeffDiscr.AP_COEFF_INTERVAL:
            coeff = type(self)(0, 0)
        else:  # self.coeff.contents.discr == CoefficientDiscr.AP_COEFF_SCALAR
            coeff = type(self)(0)
        libapron.ap_coeff_neg(coeff, self)
        return coeff


libapron.ap_coeff_alloc.argtypes = [c_uint]
libapron.ap_coeff_alloc.restype = POINTER(Coeff)
libapron.ap_coeff_set.argtypes = [POINTER(Coeff), POINTER(Coeff)]
libapron.ap_coeff_set_scalar.argtypes = [PyCoeff, PyScalar]
libapron.ap_coeff_set_interval.argtypes = [PyCoeff, PyInterval]
libapron.ap_coeff_cmp.argtypes = [PyCoeff, PyCoeff]
libapron.ap_coeff_neg.argtypes = [PyCoeff, PyCoeff]


class PyScalarCoeff(PyCoeff, metaclass=ABCMeta):

    def __init__(self, scalar: PyScalar):
        super().__init__(discr=CoeffDiscr.AP_COEFF_SCALAR, value=scalar)


class PyDoubleScalarCoeff(PyScalarCoeff):

    def __init__(self, value=0.0):
        if isinstance(value, (Coeff, PyDoubleScalar)):
            super().__init__(value)
        else:
            super().__init__(PyDoubleScalar(value))


class PyMPQScalarCoeff(PyScalarCoeff):

    def __init__(self, value_or_numerator=0, denumerator=1):
        if isinstance(value_or_numerator, (Coeff, PyMPQScalar)):
            super().__init__(value_or_numerator)
        else:
            super().__init__(PyMPQScalar(value_or_numerator, denumerator))


class PyMPFRScalarCoeff(PyScalarCoeff):

    def __init__(self, value, rounding: Rnd = Rnd.MPFR_RNDN):
        if isinstance(value, (Coeff, PyMPFRScalar)):
            super().__init__(value)
        else:
            super().__init__(PyMPFRScalar(value, rounding))


class PyIntervalCoeff(PyCoeff, metaclass=ABCMeta):

    def __init__(self, interval: PyInterval):
        super().__init__(discr=CoeffDiscr.AP_COEFF_INTERVAL, value=interval)


class PyDoubleIntervalCoeff(PyIntervalCoeff):

    def __init__(self, value_or_inf=0.0, sup=0.0):
        if isinstance(value_or_inf, (Coeff, PyDoubleInterval)):
            super().__init__(value_or_inf)
        else:
            super().__init__(PyDoubleInterval(value_or_inf, sup))


class PyMPQIntervalCoeff(PyIntervalCoeff):

    def __init__(self, value_or_inf_num=0, sup_num=0, inf_den=1, sup_den=1):
        if isinstance(value_or_inf_num, (Coeff, PyMPQInterval)):
            super().__init__(value_or_inf_num)
        else:
            super().__init__(PyMPQInterval(value_or_inf_num, sup_num, inf_den, sup_den))


class PyMPFRIntervalCoeff(PyIntervalCoeff):

    def __init__(self, value_or_inf, sup=None, rounding: Rnd = Rnd.MPFR_RNDN):
        if isinstance(value_or_inf, (Coeff, PyMPFRInterval)):
            super().__init__(value_or_inf)
        else:
            super().__init__(PyMPFRInterval(value_or_inf, sup, rounding))



"""
APRON Scalar Numbers
====================

:Author: Caterina Urban
"""
from abc import ABCMeta
from copy import deepcopy
from ctypes import *
from enum import IntEnum

from apronpy.mpfr import MPFR, PyMPFR, Rnd
from apronpy.mpq import PyMPQ, MPQ
from apronpy.cdll import libapron


class ScalarDiscr(IntEnum):
    """
    typedef enum ap_scalar_discr_t {
      AP_SCALAR_DOUBLE, /* double-precision floating-point number */
      AP_SCALAR_MPQ,    /* GMP arbitrary precision rational */
      AP_SCALAR_MPFR,   /* MPFR floating-point number */
    } ap_scalar_discr_t;
    """
    AP_SCALAR_DOUBLE = 0
    AP_SCALAR_MPQ = 1
    AP_SCALAR_MPFR = 2


class Scalar(Structure):
    """
    typedef struct ap_scalar_t {
      ap_scalar_discr_t discr;
      union {
        double dbl;
        mpq_ptr mpq; /* +infty coded by 1/0, -infty coded by -1/0 */
        mpfr_ptr mpfr;
      } val;
    } ap_scalar_t;
    """

    class Val(Union):
        """
        union {
          double dbl;
          mpq_ptr mpq; /* +infty coded by 1/0, -infty coded by -1/0 */
          mpfr_ptr mpfr;
        } val;
        """

        _fields_ = [
            ('dbl', c_double),
            ('mpq_ptr', POINTER(MPQ)),
            ('mpfr_ptr', POINTER(MPFR))
        ]

    _fields_ = [
        ('discr', c_uint),
        ('val', Val)
    ]

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        result = libapron.ap_scalar_alloc_set(byref(self)).contents
        memodict[id(self)] = result
        return result

    def __repr__(self):
        if self.discr == ScalarDiscr.AP_SCALAR_MPQ:
            return '{}'.format(self.val.mpq_ptr.contents)
        elif self.discr == ScalarDiscr.AP_SCALAR_MPFR:
            return '{}'.format(self.val.mpfr_ptr.contents)
        else:  # self.discr == Discr.AP_SCALAR_DOUBLE
            return '{}'.format(self.val.dbl)


class PyScalar(metaclass=ABCMeta):

    def __init__(self, value, discr: ScalarDiscr = ScalarDiscr.AP_SCALAR_DOUBLE):
        if isinstance(value, Scalar):
            self.scalar = byref(value)
        elif discr == ScalarDiscr.AP_SCALAR_MPQ:
            self.scalar = libapron.ap_scalar_alloc_set_mpq(value)
        elif discr == ScalarDiscr.AP_SCALAR_MPFR:
            self.scalar = libapron.ap_scalar_alloc_set_mpfr(value)
        else:
            assert discr == ScalarDiscr.AP_SCALAR_DOUBLE
            self.scalar = libapron.ap_scalar_alloc_set_double(value)

    @classmethod
    def init_infty(cls, sign: int):
        scalar = cls(0)
        libapron.ap_scalar_set_infty(scalar, sign)
        return scalar

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        result = type(self)(deepcopy(self.scalar.contents))
        memodict[id(self)] = result
        return result

    def __del__(self):
        libapron.ap_scalar_free(self)
        del self.scalar

    @property
    def _as_parameter_(self):
        return self.scalar

    @staticmethod
    def from_param(argument):
        assert isinstance(argument, PyScalar)
        return argument

    def __repr__(self):
        return '{}'.format(self.scalar.contents)

    def infty(self):
        """-1: -infty, 0: finite; 1: +infty"""
        return libapron.ap_scalar_infty(self)

    def __lt__(self, other: 'PyScalar'):
        assert isinstance(other, PyScalar)
        return libapron.ap_scalar_cmp(self, other) < 0

    def __le__(self, other: 'PyScalar'):
        assert isinstance(other, PyScalar)
        return self.__lt__(other) or self.__eq__(other)

    def __eq__(self, other: 'PyScalar'):
        assert isinstance(other, PyScalar)
        return libapron.ap_scalar_cmp(self, other) == 0

    def __ne__(self, other: 'PyScalar'):
        assert isinstance(other, PyScalar)
        return not self.__eq__(other)

    def __ge__(self, other: 'PyScalar'):
        assert isinstance(other, PyScalar)
        return self.__gt__(other) or self.__eq__(other)

    def __gt__(self, other: 'PyScalar'):
        assert isinstance(other, PyScalar)
        return libapron.ap_scalar_cmp(self, other) > 0

    def sign(self):
        """-1: negative, 0: null, +1: positive"""
        return libapron.ap_scalar_sgn(self)

    def __neg__(self) -> 'PyScalar':
        scalar = type(self)(0)
        libapron.ap_scalar_neg(scalar, self)
        return scalar


libapron.ap_scalar_alloc.restype = POINTER(Scalar)
libapron.ap_scalar_set.argtypes = [POINTER(Scalar), PyScalar]
libapron.ap_scalar_alloc_set.argtypes = [POINTER(Scalar)]
libapron.ap_scalar_alloc_set.restype = POINTER(Scalar)
libapron.ap_scalar_alloc_set_double.argtypes = [c_double]
libapron.ap_scalar_alloc_set_double.restype = POINTER(Scalar)
libapron.ap_scalar_alloc_set_mpq.argtypes = [PyMPQ]
libapron.ap_scalar_alloc_set_mpq.restype = POINTER(Scalar)
libapron.ap_scalar_alloc_set_mpfr.argtypes = [PyMPFR]
libapron.ap_scalar_alloc_set_mpfr.restype = POINTER(Scalar)
libapron.ap_scalar_set_infty.argtypes = [PyScalar, c_int]
libapron.ap_scalar_free.argtypes = [PyScalar]
libapron.ap_scalar_infty.argtypes = [PyScalar]
libapron.ap_scalar_cmp.argtypes = [PyScalar, PyScalar]
libapron.ap_scalar_sgn.argtypes = [PyScalar]
libapron.ap_scalar_neg.argtypes = [PyScalar, PyScalar]


class PyDoubleScalar(PyScalar):

    def __init__(self, value=0.0):
        if isinstance(value, Scalar):
            super().__init__(discr=ScalarDiscr.AP_SCALAR_DOUBLE, value=value)
        elif isinstance(value, c_double):
            super().__init__(discr=ScalarDiscr.AP_SCALAR_DOUBLE, value=value)
        else:
            assert isinstance(value, (int, float))
            super().__init__(discr=ScalarDiscr.AP_SCALAR_DOUBLE, value=c_double(value))


class PyMPQScalar(PyScalar):

    def __init__(self, value_or_numerator=0, denumerator=1):
        if isinstance(value_or_numerator, (Scalar, PyMPQ)):
            super().__init__(discr=ScalarDiscr.AP_SCALAR_MPQ, value=value_or_numerator)
        elif isinstance(value_or_numerator, float):
            mpq = PyMPQ(value_or_numerator)
            super().__init__(discr=ScalarDiscr.AP_SCALAR_MPQ, value=mpq)
        else:
            assert isinstance(value_or_numerator, int) and isinstance(denumerator, int)
            mpq = PyMPQ(value_or_numerator, denumerator)
            super().__init__(discr=ScalarDiscr.AP_SCALAR_MPQ, value=mpq)


class PyMPFRScalar(PyScalar):

    def __init__(self, value, rounding: Rnd = Rnd.MPFR_RNDN):
        if isinstance(value, (Scalar, PyMPFR)):
            super().__init__(discr=ScalarDiscr.AP_SCALAR_MPFR, value=value)
        else:
            super().__init__(discr=ScalarDiscr.AP_SCALAR_MPFR, value=PyMPFR(value, rounding))
